#!/usr/bin/env Rscript



#################################
###       LOAD PACKAGES       ###
#################################

packages <- c('ggplot2','raster','ResistanceGA','sp','rgdal',
              'spaa','rgeos','plyr','dplyr','scales')

pack.load <- sapply(packages, function(x) {
  suppressWarnings(suppressPackageStartupMessages(require(x, character.only = T)))
})

if( any(!pack.load) ) {
  stop("Problem loading some package(s)!")
}



#################################
###      SET PARAMETERS       ###
#################################

rgdal::setCPLConfigOption("GDAL_PAM_ENABLED", "FALSE") %>% invisible()
options( scipen = 999 )

# Set options
MAIN_DIR <- "/choose/your/working/directory/"
setwd(MAIN_DIR)
THREADS = 30
INDIR <- "./Resistance/"
POINTDIR <- "./Point/" # Available upon reasonable request
GENDIR <- "./Fst/"
OUTDIR <- "./Output/"
FULLRESDIR <- "./FullRes/"
RESULTSDIR <- "./Results/"
RASTER.NM <- c("Slope","Veg")
BARRIER.NM <- "Ridge"
RES <- "30m"
FACT <- 10
TYPE <- "single" # 'single' or 'multiple'
CAT2CONT <- TRUE
RES.AGG <- gsub("m", "", RES) %>% as.numeric() %>% prod(., FACT)
if( dir.exists(INDIR) ) system(paste0("rm -r ", INDIR))
if( dir.exists(OUTDIR) ) system(paste0("rm -r ", OUTDIR))
dir.create(INDIR)
dir.create(OUTDIR)
if( !dir.exists(POINTDIR) { dir.create(POINTDIR) } )
if( !dir.exists(RESULTSDIR) { dir.create(RESULTSDIR) } )

Mode <- function(x, na.rm = T) {
  if(na.rm) {x <- x[!is.na(x)]}
  ux <- unique(x)
  ux[which.max(tabulate(match(x, ux)))]
}
na_rm <- function(x) { x[!is.na(x)]}



#################################
###      SETUP RASTERS        ###
#################################

# Get resistance rasters, resample if necessary
RASTERS <- lapply(paste0("./Input/", RASTER.NM, RES, ".tif"), raster) %>% stack()
BARRIER <- lapply(paste0("./Input/", BARRIER.NM, RES, ".tif"), raster) %>% stack()
TYPES <- sapply(1:nlayers(RASTERS), function(i) {
  vals <- values(RASTERS[[i]]) %>% unique() %>% length()
  if(vals <= 15 & !CAT2CONT) { "categorical" } else { "continuous" }
})
TRANS <- lapply(TYPES, function(x) {
  if(x == "categorical") NA
  if(x == "continuous") "M"
})
if( FACT > 1 ) {
  RASTERS.AGG <-
    lapply(1:nlayers(RASTERS), function(i) {
      vals <- values(RASTERS[[i]]) %>% unique() %>% length()
      r <- aggregate(
        RASTERS[[i]], fact = FACT, fun = 
          if(TYPES[[i]] == "categorical") { Mode } else { mean })
      MIN <- values(r) %>% na_rm() %>% min()
      if( (MIN < 0) & FALSE ) {
        # Temporary experiment, for NDMI with neg. values
        ADD <- diff(c(MIN, 0))
        r <- r + ADD
      }
      return(r)
    }) %>% stack()
} else {
  RASTERS.AGG <- RASTERS
}

for(i in 1:nlayers(RASTERS.AGG)) {
  writeRaster(
    RASTERS.AGG[[i]], 
    paste0(
      INDIR, 
      RASTER.NM[[i]], 
      RES.AGG, "m"), 
    format = "ascii", overwrite = T
  )
  writeRaster(
    RASTERS[[i]], 
    paste0(
      FULLRESDIR, 
      RASTER.NM[[i]], 
      RES), 
    format = "ascii", overwrite = T
  )
}



##############################
###      SETUP POINTS      ###
##############################

# Get points
pts <- readOGR(dsn = POINTDIR, layer = "GenMdwPolys") # Available upon reasonable request
xy  <- coordinates(pts)
colnames(xy) <- c("x","y")
samples <- data.frame(pop = pts$meadowID, xy)
write.table(samples, file = paste0(POINTDIR, "samples.txt"), sep="\t", col.names=F, row.names=F)

# Convert points to Euclidean distances
p <- gCentroid(pts, byid = TRUE)
XY <- p %>%
  spDists() %>%
  matrix(
    byrow = T, nrow = nrow(.), 
    dimnames = 
      list(pts@data$meadowID, pts@data$meadowID)) %>%
  as.dist() %>% 
  dist2list() %>%
  tibble() %>%
  dplyr::mutate(col = as.integer(as.character(col))) %>%
  dplyr::mutate(row = as.integer(as.character(row))) %>%
  dplyr::filter(col < row) %>%
  dplyr::rename(pop1 = col, pop2 = row, EucDist = value) %>%
  dplyr::arrange(pop1, pop2)



###########################
###      SETUP Fst      ###
###########################

# Get genetic distance
response <- read.table(
  paste0(GENDIR, "./Fst.txt"), 
  header = T, sep = "\t", stringsAsFactors = F) %>% 
  tibble() %>%
  dplyr::select(pop1, pop2, fst_wc.mean) %>%
  dplyr::mutate(pop1 = as.integer(pop1)) %>%
  dplyr::mutate(pop2 = as.integer(pop2)) %>%
  dplyr::arrange(pop1, pop2)

# Check that data are aligned correctly, should be a geogenetic correlation
XY. <- XY %>% dplyr::left_join(x = XY, y = response, by = c("pop1","pop2"))
LM <- lm(fst_wc.mean ~ EucDist + (pop1 | 1), data = XY.) %>% summary()
LM$adj.r.squared



####################################
###   SETUP / RUN RESISTANCEGA   ###
####################################

# Create R object with genetic algorithm optimization settings
GA.inputs <- GA.prep(
  ASCII.dir = INDIR,
  Results.dir = RESULTSDIR,
  select.trans = TRANS,
  max.cat = 1E6,
  max.cont = 1E6,
  cont.shape = "increase",
  seed = 12345,
  # gaisl = TRUE,
  run = 10,
  maxiter = 50,
  opt.digits = 1,
  parallel = THREADS
)

# Prepare data for optimization using gdistance
gdist.inputs <- gdist.prep(
  n.Pops = nrow(samples),
  samples = p,
  response = response$fst_wc.mean,
  method = 'costDistance',
  directions = 8,
  transitionFunction = function(x) 1/mean(x)
)

if( TYPE == "single" ) {
  # Run single surface optimization
  RESULT <- SS_optim(
    gdist.inputs = gdist.inputs,
    GA.inputs = GA.inputs
  )
  write.table(RESULT$ContinuousResults, "./Output/Result.txt", row.names = F, sep = "\t", quote = F)
  write.table(RESULT$AICc, "./Output/AICc.txt", row.names = F, sep = "\t", quote = F)
  write.table(RESULT$k, "./Output/k.txt", row.names = F, sep = "\t", quote = F)
  qc <- lapply(1:length(RESULT$cd), function(i) {
    write.table(RESULT$cd[[i]], paste0("./Output/CD_", names(RESULT$cd)[i], ".txt"), 
                row.names = F, sep = "\t", quote = F)
    return(0)
  })
  qc <- lapply(1:length(RESULT$MLPE.list), function(i) {
    x <- RESULT$MLPE.list[[i]]
    save(x, file = paste0("./Output/MLPE_", names(RESULT$MLPE.list)[i], ".RData"))
    return(0)
  })
  qc <- lapply(1:length(RESULT$ga), function(i) {
    x <- summary(RESULT$ga[[i]])
    save(x, file = paste0("./Output/GA_", names(RESULT$MLPE.list)[i], ".RData"))
    return(0)
  })
}

if( TYPE == "multiple" ) {
  # Run multiple surface optimization
  RESULT2 <- MS_optim(
    gdist.inputs = gdist.inputs,
    GA.inputs = GA.inputs
  )
  RESULT2$ContinuousResults <- data.frame(
    Equation = character(0),
    shape = numeric(0),
    max = numeric(0)
  )
  MX.CAT <- sapply(1:nlayers(RASTERS), function(i) {
    if(TYPES[[i]] == "categorical") RASTERS[[i]] %>% values() %>% table() %>% length()
    else NA
  }) %>% max(na.rm = T)
  col.nms <- paste0("Feature", 1:MX.CAT)
  RESULT2$CategoricalResults <- data.frame()
  for (i in 1:MX.CAT){
    RESULT2$CategoricalResults[,col.nms[i]] <- numeric(0)
  }
}

if( TYPE == "multiple" ) {
  
  # Combine surfaces into one composite surface
  PARM <- RESULT2$GA.summary@solution
  GA.inputs.orig <- GA.inputs
  GA.inputs.orig$ASCII.dir <- FULLRESDIR
  GA.inputs.orig$Resistance.stack <- RASTERS
  Resist <- Combine_Surfaces(
    PARM = PARM,
    gdist.inputs = gdist.inputs,
    GA.inputs = GA.inputs,
    out = OUTDIR,
    rescale = TRUE,
    p.contribution = TRUE
  )
  
  # Reorganize output to be like single-surface output
  I <- 1
  for( t in 1:length(TYPES) ) {
    if( TYPES[t] == "continuous" ) {
      RESULT2$ContinuousResults <- rbind(
        RESULT2$ContinuousResults,
        data.frame(
          Equation = RESULT2$GA.summary@solution[I],
          shape = RESULT2$GA.summary@solution[I+1],
          max = RESULT2$GA.summary@solution[I+2]
        )
      )
      I <- I + 3
    }
    if( TYPES[t] == "categorical" ) {
      MX.CAT.t <- RASTERS[[t]] %>% values() %>% table() %>% length()
      J <- I + MX.CAT.t - 1
      df <- t(RESULT2$GA.summary@solution[I:J]) %>% data.frame()
      add <- MX.CAT - MX.CAT.t
      if( add ) df <- cbind(df, t(rep(NA, add)))
      names(df) <- col.nms
      RESULT2$CategoricalResults <- rbind(
        RESULT2$CategoricalResults, df
      )
    }
  }
}

if( TYPE == "single" ) RESULT. <- RESULT
if( TYPE == "multiple" ) RESULT. <- RESULT2

for( t in 1:length(TYPES) ) {
  if( TYPES[t] == "continuous" ) {
    # Grab continuous trans type, shape, and max
    TRANS <- RESULT.$ContinuousResults$Equation[t]
    SHAPE <- RESULT.$ContinuousResults$shape[t]
    MAX <- RESULT.$ContinuousResults$max[t]
    t.range <- range(values(RASTERS[[t]]), na.rm = T) #%>% round_any(., 10, ceiling)
    plot.t1 <- Plot.trans(
      PARM = c(SHAPE, MAX),
      Resistance =  t.range,
      transformation = TRANS
    )
    plot.t1 <- plot.t1 +
      scale_y_continuous(label = comma) +
      labs(title = RASTER.NM[t]) +
      theme(plot.title = element_text(hjust = 0.5, size = 16, face = "bold")) %>%
      suppressMessages()
    ggsave(
      paste0("./Output/", RASTER.NM[t], "_vals.pdf"),
      plot = plot.t1,
      width = 8,
      height = 6,
      units = "in"
    )
    vals <- switch(t,
      seq(0, 90, 10),
      seq(1, 10, 1),
    )
    EQ <- ResistanceGA:::get.EQ(TRANS)
    parm <- c(EQ, SHAPE, MAX)
    new.vals <- switch(EQ,
      ResistanceGA:::Inv.Rev.Monomolecular(vals, parm),
      ResistanceGA:::Inv.Rev.Ricker(vals, parm),
      ResistanceGA:::Monomolecular(vals, parm),
      ResistanceGA:::Ricker(vals, parm),
      ResistanceGA:::Rev.Monomolecular(vals, parm),
      ResistanceGA:::Rev.Ricker(vals, parm),
      ResistanceGA:::Inv.Monomolecular(vals, parm),
      ResistanceGA:::Inv.Ricker(vals, parm)
    )
    new.vals.df <- tibble(
      Value = vals,
      Transformed = new.vals
    )
    write.table(new.vals.df, paste0("./Output/", RASTER.NM[t], ".txt"), 
                sep = "\t", row.names = F, quote = F)
  }
  
  if( TYPES[t] == "categorical" ) {
    # Grab categorical trans df
    cat.ori <- RASTERS[[t]] %>% values() %>% table() %>% names()
    RESULT.$CategoricalResults$Surface <- paste0(RASTER.NM[[t]], RES.AGG, "m")
    CAT <- RESULT.$CategoricalResults %>% 
      dplyr::select(., Surface, contains("Feature")) %>% 
      tidyr::pivot_longer(!Surface, names_to = "Feature", values_to = "Transformed") %>%
      dplyr::mutate(Original = cat.ori) %>%
      dplyr::select(Feature, Original, Transformed)
    print(CAT)
  }
}



##########################################
###   TRANSFORM RASTER USING RESULTS   ###
##########################################

# Transform original resolution continuous raster
idx.na <- lapply(RASTERS, function(x) is.na(values(x))) %>% Reduce(`|`, .) %>% which()
RASTERS.TRANS <- lapply(which(TYPES == "continuous"), function(t) {
  TRANS <- RESULT.$ContinuousResults$Equation[t]
  SHAPE <- RESULT.$ContinuousResults$shape[t]
  MAX <- RESULT.$ContinuousResults$max[t]
  max.val <- max(values(RASTERS[[t]]), na.rm = T)
  if(max.val != 10) {
    R <- RASTERS[[t]] * (10 / max.val)
  } else {
    R <- RASTERS[[t]]
  }
  r.tran <- Resistance.tran(
    transformation = TRANS,
    shape = SHAPE,
    max = MAX,
    r = R
  )
  r.tran[BARRIER == 1E6] <- 1E6
  r.tran[idx.na] <- NA
  return(r.tran)
})

plot(RASTERS.TRANS[[1]])
plot(RASTERS.TRANS[[2]])



#################################
###  CREATE WEIGHTED COMBOS   ###
#################################

proportions <- tibble(
  x = seq(0, 10, 1),
  y = rev(seq(0, 10, 1))
) / 10

for( p in 1:nrow(proportions) ) {
  p1 <- proportions[p,1]
  p2 <- proportions[p,2]
  R <- RASTERS.TRANS[[1]] * p1 + RASTERS.TRANS[[2]] * p2
  print(paste0(p1, " and ", p2))
  print(range(values(R), na.rm = T))
  plot(R)
  writeRaster(R, paste0(OUTDIR, "SlopeVeg_", p1*100, "_", p2*100, ".tif"), overwrite = T)
}




